# type: ignore

from ..sync import to_sync


def patch_sync_apis():
    from openai.version import VERSION

    if VERSION.startswith("0"):
        from promplate.llm.openai import v0

        v0.TextComplete.__call__ = to_sync(v0.AsyncTextComplete.__call__)
        v0.TextGenerate.__call__ = to_sync(v0.AsyncTextGenerate.__call__)
        v0.ChatComplete.__call__ = to_sync(v0.AsyncChatComplete.__call__)
        v0.ChatGenerate.__call__ = to_sync(v0.AsyncChatGenerate.__call__)

    else:
        from promplate.llm.openai import v1

        v1.SyncTextOpenAI.complete = v1.TextComplete.__call__ = to_sync(v1.AsyncTextComplete.__call__)
        v1.SyncTextOpenAI.generate = v1.TextGenerate.__call__ = to_sync(v1.AsyncTextGenerate.__call__)
        v1.SyncChatOpenAI.complete = v1.ChatComplete.__call__ = to_sync(v1.AsyncChatComplete.__call__)
        v1.SyncChatOpenAI.generate = v1.ChatGenerate.__call__ = to_sync(v1.AsyncChatGenerate.__call__)



try:
    from js import Promise  # type: ignore
    from pyodide.ffi import run_sync  # type: ignore

    def stack_switching_supported():  # type: ignore
        try:
            run_sync(Promise.resolve())
            return True
        except RuntimeError:
            return False

except ImportError:

    def stack_switching_supported():
        return False



from contextlib import suppress
from os import getenv
from typing import cast

from pyodide.code import run_js
from pyodide.ffi import JsCallable, register_js_module

from ..proxy import to_js

get_openai_js_script = """(
  async () => {
    const [{ OpenAI }, version] = await Promise.all([import("openai"), import("openai/version")])
    return { OpenAI, version, __all__: [] }
  }
)()"""


async def ensure_openai(fallback_import_url: str):
    with suppress(ModuleNotFoundError):
        import openai.version

        return

    openai = await run_js(get_openai_js_script.replace("openai", fallback_import_url))

    register_js_module("openai", openai)


def translate_openai():
    from openai import OpenAI

    js_openai_class = cast(JsCallable, OpenAI)

    def make_client(
        api_key: str | None = None,
        organization: str | None = None,
        base_url: str | None = None,
        timeout: float | None = None,
        max_retries: int | None = None,
        default_headers: dict | None = None,
        default_query: dict | None = None,
        **_,
    ):
        return js_openai_class.new(
            apiKey=api_key or getenv("OPENAI_API_KEY", ""),
            organization=organization or getenv("OPENAI_ORG_ID"),
            baseURL=base_url or getenv("OPENAI_BASE_URL"),
            timeout=timeout,
            maxRetries=max_retries,
            defaultHeaders=to_js(default_headers),
            defaultQuery=to_js(default_query),
            dangerouslyAllowBrowser=True,
        )

    return make_client



from pyodide.ffi import JsProxy


def to_js(obj):
    from js import Object  # type: ignore
    from pyodide.ffi import to_js

    return to_js(obj, dict_converter=Object.fromEntries)


def to_py(obj: JsProxy):
    return obj.to_py() if isinstance(obj, JsProxy) else obj



class WarnPrinter:
    def __init__(self, warn: str):
        self.warn = warn

    def __call__(self, *args, **kwargs):
        return self

    def __await__(self):
        return self

    def __repr__(self):
        return self.warn

    def __getattr__(self, _):
        return self

    def __getitem__(self, _):
        return self


sync_api_warning = WarnPrinter("Synchronous APIs are not available in pyodide unless stack switching is supported.")



from pyodide.ffi import JsProxy


def is_fake_openai():
    import openai

    return isinstance(openai, JsProxy)


def is_fake_httpx():
    try:
        import httpx
    except ModuleNotFoundError:
        return True
    else:
        return isinstance(httpx, JsProxy)



from inspect import isawaitable
from typing import AsyncIterable, Awaitable, Callable, Iterable, overload

from .stack_switching import stack_switching_supported
from .warn import sync_api_warning

STOP_ITERATION = object()


def syncify[T](iterable: AsyncIterable[T]) -> Iterable[T]:
    from pyodide.ffi import run_sync

    it = aiter(iterable)
    while True:
        res = run_sync(anext(it, STOP_ITERATION))
        if res is STOP_ITERATION:
            break
        else:
            yield res  # type: ignore


@overload
def to_sync[**P, T](func: Callable[P, AsyncIterable[T]]) -> Callable[P, Iterable[T]]: ...
@overload
def to_sync[**P, T](func: Callable[P, Awaitable[T]]) -> Callable[P, T]: ...


def to_sync(func):  # type: ignore
    from functools import wraps

    from pyodide.ffi import run_sync

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not stack_switching_supported():
            return sync_api_warning
        res = func(*args, **kwargs)
        return run_sync(res) if isawaitable(res) else syncify(res)

    return wrapper



from contextlib import suppress
from functools import wraps
from inspect import isclass, isfunction
from pathlib import Path
from types import FunctionType

from .utils.check import is_fake_httpx, is_fake_openai
from .utils.stack_switching import stack_switching_supported


def patch_promplate():
    import promplate

    if is_fake_httpx():
        from promplate.prompt.template import Loader

        @classmethod
        @wraps(Loader.afetch)
        async def afetch(cls, url: str, **kwargs):
            from pyodide.http import pyfetch

            res = await pyfetch(cls._join_url(url), **kwargs)
            obj = cls(await res.text())
            obj.name = Path(url).stem

            return obj

        @classmethod
        @wraps(Loader.fetch)
        def fetch(cls, url: str, **kwargs):
            if stack_switching_supported():
                from pyodide.ffi import run_sync  # type: ignore

                return run_sync(cls.afetch(url, **kwargs))

            else:
                from pyodide.http import open_url

                res = open_url(cls._join_url(url))
                obj = cls(res.read())
                obj.name = Path(url).stem

                return obj

        Loader.afetch = afetch  # type: ignore
        Loader.fetch = fetch  # type: ignore

    if not is_fake_openai():
        return

    patch_class(promplate.node.Interruptable)

    from .utils.proxy import to_js

    with suppress(ModuleNotFoundError):
        import promplate.llm.openai as o

        def patched_ensure(text_or_list: list[promplate.Message] | str):
            """This function is patched to return a JS array. So it should not be called from Python."""
            return to_js(promplate.prompt.chat.ensure(text_or_list))

        def patch_function(func: FunctionType):
            from promplate.prompt.template import SafeChainMapContext as ChainMap

            func = FunctionType(func.__code__, ChainMap({"ensure": patched_ensure}, func.__globals__), func.__name__, func.__defaults__, func.__closure__)

            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*(to_js(i) for i in args), **{i: to_js(j) for i, j in kwargs.items()})

            return wrapper

        for cls in o.v1.__dict__.values():
            if isclass(cls):
                for name, func in [*cls.__dict__.items()]:
                    if isfunction(func) and func.__name__ == "__call__":
                        setattr(cls, name, patch_function(func))

        from .utils.openai.sync import patch_sync_apis

        patch_sync_apis()


async def patch_openai(fallback_import_url: str = "https://esm.sh/openai"):
    from .utils.openai import ensure_openai, translate_openai

    await ensure_openai(fallback_import_url)

    if not is_fake_openai():
        return

    import openai

    openai.Client = openai.AsyncClient = translate_openai()


def patch_httpx():
    import promplate.llm.openai.v1 as o

    o._get_client = o._get_aclient = lambda: None  # type: ignore


async def patch_all():
    await patch_openai()
    patch_httpx()
    patch_promplate()


def patch_params(func: FunctionType):
    from .utils.proxy import to_py

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*(to_py(i) for i in args), **{i: to_py(j) for i, j in kwargs.items()})

    return wrapper


def patch_class(cls: type):
    for name, func in cls.__dict__.items():
        if isfunction(func):
            setattr(cls, name, patch_params(func))
